package com.flow.common.core.aspect;

import com.flow.common.core.annotation.RateLimiter;
import com.flow.common.core.enums.LimitType;
import com.flow.common.core.exception.BaseException;
import com.flow.common.core.util.RequestUtils;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Before;
import org.aspectj.lang.reflect.MethodSignature;
import org.redisson.api.RRateLimiter;
import org.redisson.api.RateIntervalUnit;
import org.redisson.api.RateType;
import org.redisson.api.RedissonClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.lang.reflect.Method;

/**
 * 限流处理
 */
@Aspect
@Component
public class RateLimiterAspect {
    private final Logger log = LoggerFactory.getLogger(RateLimiterAspect.class);

    @Autowired
    private RedissonClient redissonClient;

    @Before(value = "@annotation(rateLimiter)")
    public void before(JoinPoint joinPoint, RateLimiter rateLimiter) {
        int rateInterval = rateLimiter.time();
        RateIntervalUnit unit = rateLimiter.unit();
        int rate = rateLimiter.count();
        String combineKey = getCombineKey(rateLimiter, joinPoint);
        try {
            RateType rateType = RateType.OVERALL;
            if (rateLimiter.limitType() == LimitType.CLUSTER) {
                rateType = RateType.PER_CLIENT;
            }
            RRateLimiter rRateLimiter = redissonClient.getRateLimiter(combineKey);
            rRateLimiter.setRate(rateType, rate, rateInterval, unit);
            if (rRateLimiter.tryAcquire()) {
                long number = rRateLimiter.availablePermits();
                log.info("限流令牌 => {},剩余令牌 => {},缓存Key => {}", rate, number, combineKey);
            } else {
                throw new BaseException("请求太频繁了，请稍后再试");
            }
        } catch (Exception e) {
            throw new BaseException("服务器限流异常，请稍后重试");
        }
    }

    public String getCombineKey(RateLimiter rateLimiter, JoinPoint joinPoint) {
        StringBuilder stringBuffer = new StringBuilder(rateLimiter.key());
        if (rateLimiter.limitType() == LimitType.IP) {
            String ip = RequestUtils.getIp();
            stringBuffer.append("ip:").append(ip);
        }
        MethodSignature signature = (MethodSignature) joinPoint.getSignature();
        Method method = signature.getMethod();
        Class<?> targetClass = method.getDeclaringClass();
        stringBuffer.append("#").append(targetClass.getName()).append(".").append(method.getName()).append("()");
        return stringBuffer.toString();
    }
}
