package com.flow.service.impl;


import com.baomidou.mybatisplus.core.toolkit.CollectionUtils;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.flow.common.core.exception.BaseException;
import com.flow.common.core.model.PageResult;
import com.flow.common.core.util.SecurityContextUtil;
import com.flow.common.mybatis.constant.SqlConstant;
import com.flow.common.mybatis.service.impl.BaseServiceImpl;
import com.flow.dao.UserDao;
import com.flow.entity.Role;
import com.flow.entity.StorageFile;
import com.flow.entity.User;
import com.flow.mapstruct.UserMapper;
import com.flow.model.UserQuery;
import com.flow.service.FileService;
import com.flow.service.UserService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import java.io.IOException;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

@Service
public class UserServiceImpl extends BaseServiceImpl<UserDao, User> implements UserService {
    @Autowired
    private UserDao userDao;
    @Autowired
    private PasswordEncoder passwordEncoder;
    @Autowired
    private UserMapper userMapper;
    @Autowired
    private FileService fileService;

    @Override
    public List<User> search(UserQuery userQuery) {
        User user = userMapper.toEntity(userQuery);
        return userDao.lambdaQueryChain()
                .setEntity(user)
                .list();
    }

    @Override
    public PageResult<User> getList(UserQuery userQuery) {
        User user = userMapper.toEntity(userQuery);
        Page<User> page = userDao.lambdaQueryChain()
                .setEntity(user)
                .page(
                        new Page<>(userQuery.getPage(), userQuery.getLimit())
                );
        return new PageResult<>(page.getTotal(), page.getRecords());
    }

    @Override
    public List<Role> getRoles(String userId) {
        return userDao.getRoles(userId);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void authRoles(String userId, Set<String> roleIds) {
        userDao.delNotExistRoles(userId, roleIds);
        if (CollectionUtils.isNotEmpty(roleIds)) {
            userDao.authRoles(userId, roleIds);
        }
    }

    @Override
    public List<User> getOptions(Set<String> userIds) {
        return userDao.lambdaQueryChain()
                .in(User::getUsername, userIds)
                .list();
    }

    @Override
    public User getByUsername(String username) {
        return userDao.lambdaQueryChain()
                .eq(User::getUsername, username)
                .last(SqlConstant.LIMIT1)
                .one();
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public User create(User user) {
        if (user.getUsername().equals(user.getLeader())) {
            throw new BaseException("上级领导不能是自己");
        }
        Optional<User> optional = userDao.lambdaQueryChain()
                .eq(User::getUsername, user.getUsername())
                .oneOpt();
        if (optional.isPresent()) {
            throw new BaseException("用户名已存在");
        }
        user.setPassword(passwordEncoder.encode("123456"));
        userDao.insert(user);
        Set<String> roleIds = user.getRoleIds();
        if (CollectionUtils.isNotEmpty(roleIds)) {
            this.authRoles(user.getUsername(), roleIds);
        }
        return user;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public User update(User user) {
        if (user.getUsername().equals(user.getLeader())) {
            throw new BaseException("上级领导不能是自己");
        }
        userDao.updateById(user);
        Set<String> roleIds = user.getRoleIds();
        if (CollectionUtils.isNotEmpty(roleIds)) {
            this.authRoles(user.getUsername(), roleIds);
        }
        return user;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void delete(Set<Long> ids) {
        userDao.deleteBatchIds(ids);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public StorageFile updateAvatar(MultipartFile multipartFile) throws IOException {
        String contentType = multipartFile.getContentType();
        if (contentType == null || !contentType.startsWith("image")) {
            throw new BaseException("上传的文件不是图片");
        }
        StorageFile storageFile = fileService.upload(multipartFile);
        userDao.lambdaUpdateChain()
                //.set(User::getAvatar, String.format("/api%s", storageFile.getUrl()))
                .set(User::getAvatar, storageFile.getUrl())
                .eq(User::getUsername, SecurityContextUtil.getUserId())
                .update();
        return storageFile;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void resetPassword(String username) {
        User user = userDao.lambdaQueryChain()
                .eq(User::getUsername, username)
                .oneOpt()
                .orElseThrow(() -> new BaseException("用户不存在"));
        user.setPassword(passwordEncoder.encode("123456"));
        userDao.updateById(user);
    }

    @Override
    public User getLeader(String username) {
        return userDao.getLeader(username);
    }

    @Override
    public User getLeader(String username, Integer level) {
        User leader = this.getLeader(username);
        if (Objects.nonNull(leader) && level > 1) {
            return getLeader(leader.getUsername(), level - 1);
        } else {
            return leader;
        }
    }

    @Override
    public User getOrgLeader(String username) {
        return userDao.getOrgLeader(username);
    }

    @Override
    public User getOrgLeader(String username, Integer level) {
        User orgLeader = this.getOrgLeader(username);
        if (Objects.nonNull(orgLeader) && level > 1) {
            return getLeader(orgLeader.getUsername(), level - 1);
        } else {
            return orgLeader;
        }
    }
}
