package com.flow.service.impl;

import com.baomidou.mybatisplus.core.toolkit.CollectionUtils;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.flow.common.core.exception.BaseException;
import com.flow.common.core.model.OauthUserDetails;
import com.flow.common.core.util.SecurityContextUtil;
import com.flow.common.core.util.StrUtil;
import com.flow.common.mybatis.service.impl.BaseServiceImpl;
import com.flow.common.redis.service.RedisService;
import com.flow.dao.FlowDefineDao;
import com.flow.entity.FlowDefine;
import com.flow.entity.FormModel;
import com.flow.entity.form.Field;
import com.flow.entity.form.serial.SerialRule;
import com.flow.entity.node.*;
import com.flow.entity.settings.Settings;
import com.flow.entity.settings.TitleConfig;
import com.flow.enums.DuplicateApprovalEnum;
import com.flow.enums.ProcessStatus;
import com.flow.model.FormInfo;
import com.flow.model.StartProcess;
import com.flow.service.FlowDefineService;
import com.google.common.collect.Lists;
import org.apache.commons.lang3.StringUtils;
import org.flowable.common.engine.impl.identity.Authentication;
import org.flowable.engine.RepositoryService;
import org.flowable.engine.RuntimeService;
import org.flowable.engine.repository.ProcessDefinition;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.CachePut;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.cache.annotation.Caching;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

@Service
public class FlowDefineServiceImpl extends BaseServiceImpl<FlowDefineDao, FlowDefine> implements FlowDefineService {
    @Autowired
    private FlowDefineDao flowDefineDao;
    @Autowired
    private RepositoryService repositoryService;
    @Autowired
    private RuntimeService runtimeService;
    @Autowired
    private FlowDefineService _this;
    @Autowired
    private RedisService<Long> redisService;
    @Autowired
    private RedisService<Map<String, Object>> redisMapService;
    @Autowired
    private ObjectMapper objectMapper;
    private static final String TEMPORARILY_CACHE_PREFIX = "flow:temp:";

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void startProcess(StartProcess startProcess) {
        FlowDefine flowDefine = _this.getDefine(startProcess.getDefineId());
        if (Objects.isNull(flowDefine)) {
            throw new BaseException("流程不存在");
        }
        Integer suspend = flowDefine.getSuspend();
        if (suspend == 2) {
            throw new BaseException(String.format("流程（%s）已挂起，无法发起流程！", flowDefine.getName()));
        }
        Map<String, Object> values = startProcess.getValues();
        OauthUserDetails userDetails = SecurityContextUtil.getUserInfo();
        if (Objects.nonNull(userDetails)) {
            values.put("_DEPT_ID", userDetails.getDeptId());
            values.put("_LEADER", userDetails.getLeader());
            values.put("_ROLE_IDS", userDetails.getRoleIds());
        }
        values.put("initiator", SecurityContextUtil.getUserId());
        values.put("initiatorName", SecurityContextUtil.getNickName());
        values.put("_FLOWABLE_SKIP_EXPRESSION_ENABLED", true);
        values.put("_SKIP_INITIATOR", true);
        List<ApprovalNode> approvalNodes = flowDefine.findFlowNodesOfType(ApprovalNode.class);
        for (ApprovalNode approvalNode : approvalNodes) {
            values.put(String.format("%sCollection", approvalNode.getId()), approvalNode.getAssignees(values, flowDefine));
        }
        List<CcNode> ccNodes = flowDefine.findFlowNodesOfType(CcNode.class);
        for (CcNode ccNode : ccNodes) {
            values.put(String.format("%sCollection", ccNode.getId()), ccNode.getAssignees(values));
        }
        List<NotifyNode> notifyNodes = flowDefine.findFlowNodesOfType(NotifyNode.class);
        for (NotifyNode notifyNode : notifyNodes) {
            values.put(String.format("%sCollection", notifyNode.getId()), notifyNode.getAssignees(values));
        }
        List<HttpNode> httpNodes = flowDefine.findFlowNodesOfType(HttpNode.class);
        for (HttpNode httpNode : httpNodes) {
            List<String> errorUsers = httpNode.getErrorUsers();
            if (CollectionUtils.isNotEmpty(errorUsers)) {
                values.put(String.format("%sCollection", httpNode.getId()),
                        CollectionUtils.isNotEmpty(errorUsers) ? errorUsers :
                                Lists.newArrayList(SecurityContextUtil.getUserId())
                );
            }
        }
        // 生成流水号
        FormModel formModel = flowDefine.getForm();
        List<Field> fields = formModel.findByName("SerialNumber");
        for (Field field : fields) {
            String value = this.generateSerialNumber(field, flowDefine.getKey());
            if (StringUtils.isNotBlank(value)) {
                values.put(field.getId(), value);
            }
        }
        // 生成流程实例ID
        String currentDate = DateTimeFormatter.ofPattern("yyyyMMdd").format(LocalDate.now());
        String serialKey = String.format("flow:serial:%s", currentDate);
        Long increment = redisService.incr(serialKey, 1L);
        if (increment == 1L) {
            long secondsInYear = 365L * 24 * 60 * 60;
            redisService.expire(serialKey, secondsInYear);
        }
        String serial = String.format("%3s", increment).replace(" ", "0");
        String instanceId = String.format("fw%s%s", currentDate, serial);
        String instanceName = startProcess.getInstanceName();
        Settings settings = flowDefine.getSettings();
        TitleConfig titleConfig = settings.getTitle();
        if ("custom".equals(titleConfig.getType())) {
            String value = titleConfig.getValue();
            if (StringUtils.isNotBlank(value)) {
                instanceName = StrUtil.format(value, values);
            }
        }
        DuplicateApprovalEnum duplicate = settings.getDuplicate();
        if (Objects.nonNull(duplicate)) {
            values.put("_DUPLICATE_APPROVAL", duplicate.getType());
        }
        Authentication.setAuthenticatedUserId(SecurityContextUtil.getUserId());
        runtimeService.createProcessInstanceBuilder()
                .processDefinitionId(startProcess.getDefineId())
                .name(instanceName)
                .owner(SecurityContextUtil.getUserId())
                .predefineProcessInstanceId(instanceId)
                .businessStatus(ProcessStatus.RUNNING.toString())
                .variables(values)
                .start();
        // 删除暂存
        redisMapService.hdel(TEMPORARILY_CACHE_PREFIX + SecurityContextUtil.getUserId(), startProcess.getDefineId());
        try {
            runtimeService.setVariable(instanceId, "_SKIP_INITIATOR", false);
        } catch (Exception ignored) {
        }
    }

    private String generateSerialNumber(Field field, String modelKey) {
        Map<String, Object> props = field.getProps();
        Object rules = props.get("rules");
        List<SerialRule> serialRules = objectMapper.convertValue(rules, new TypeReference<List<SerialRule>>() {
        });
        String key = String.format("flow:serial:%s:%s", modelKey, field.getId());
        return serialRules.stream().map(e -> e.generate(key)).collect(Collectors.joining());
    }

    @Override
    public List<FlowDefine> list(FlowDefine flowDefine) {
        return flowDefineDao.getList(flowDefine);
    }

    @Caching(
            put = @CachePut(value = "process", key = "#defineId", unless = "#result == null"),
            evict = @CacheEvict(value = "process", key = "#defineId", condition = "#result == null")
    )
    @Override
    public FlowDefine suspend(String defineId) {
        ProcessDefinition processDefinition = this.repositoryService.getProcessDefinition(String.valueOf(defineId));
        if (processDefinition.isSuspended()) {
            this.repositoryService.activateProcessDefinitionById(String.valueOf(defineId));
            return _this.getDefine(defineId);
        } else {
            this.repositoryService.suspendProcessDefinitionById(String.valueOf(defineId));
            return null;
        }
    }

    @Cacheable(key = "#defineId", value = "process", unless = "#result == null")
    @Override
    public FlowDefine getDefine(String defineId) {
        return flowDefineDao.getByDefineId(defineId);
    }

    @Override
    public FormInfo getStartForm(String defineId) {
        FlowDefine define = _this.getDefine(defineId);
        if (Objects.isNull(define)) {
            throw new BaseException("流程不存在");
        }
        RootNode rootNode = (RootNode) define.getProcess();
        List<FieldProperty> formProperties = rootNode.getFormProperties();
        Map variables = redisMapService.hget(TEMPORARILY_CACHE_PREFIX + SecurityContextUtil.getUserId(), defineId, Map.class);
        return new FormInfo(define, formProperties, variables);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void storeFormDataTemporarily(StartProcess startProcess) {
        redisMapService.hset(TEMPORARILY_CACHE_PREFIX + SecurityContextUtil.getUserId(), startProcess.getDefineId(), startProcess.getValues());
    }

    @Override
    public Map<String, Object> getFormDataTemporarily(String defineId) {
        return redisMapService.hget(TEMPORARILY_CACHE_PREFIX + SecurityContextUtil.getUserId(), defineId, Map.class);
    }
}
