package com.flow.service.impl;

import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.flow.common.core.exception.BaseException;
import com.flow.common.mybatis.constant.SqlConstant;
import com.flow.common.mybatis.service.impl.BaseServiceImpl;
import com.flow.dao.FlowModelDao;
import com.flow.entity.*;
import com.flow.entity.form.Field;
import com.flow.entity.settings.PrintConfig;
import com.flow.entity.settings.Settings;
import com.flow.model.spreadsheet.SpreadSheet;
import com.flow.service.FlowDefineService;
import com.flow.service.FlowModelHistoryService;
import com.flow.service.FlowModelService;
import com.flow.service.ReportTemplateService;
import com.google.common.collect.Lists;
import org.apache.commons.lang3.StringUtils;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.flowable.bpmn.converter.BpmnXMLConverter;
import org.flowable.bpmn.model.BpmnModel;
import org.flowable.engine.RepositoryService;
import org.flowable.engine.repository.Deployment;
import org.flowable.engine.repository.ProcessDefinition;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletResponse;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class FlowModelServiceImpl extends BaseServiceImpl<FlowModelDao, FlowModel> implements FlowModelService {
    @Autowired
    private FlowModelDao flowModelDao;
    @Autowired
    private FlowModelHistoryService flowModelHistoryService;
    @Autowired
    private RepositoryService repositoryService;
    @Autowired
    private FlowDefineService flowDefineService;
    @Autowired
    private ObjectMapper objectMapper;
    @Autowired
    private ReportTemplateService reportTemplateService;

    @Override
    public List<FlowModel> list(FlowModel flowModel) {
        return flowModelDao.lambdaQueryChain()
                .select(FlowModel::getId,
                        FlowModel::getCode,
                        FlowModel::getName,
                        FlowModel::getIcon,
                        FlowModel::getVersion,
                        FlowModel::getSort,
                        FlowModel::getGroupId,
                        FlowModel::getEnable,
                        FlowModel::getRemark,
                        FlowModel::getAdmin,
                        FlowModel::getCreatedBy,
                        FlowModel::getUpdatedBy,
                        FlowModel::getCreateTime,
                        FlowModel::getUpdateTime)
                .list();
    }

    @Override
    public void downloadXml(FlowModel flowModel) throws IOException {
        BpmnModel bpmnModel = flowModel.toBpmnModel();
        byte[] xmlBytes = new BpmnXMLConverter().convertToXML(bpmnModel);
        if (Objects.nonNull(xmlBytes)) {
            String fileName = flowModel.getName().replaceAll(" ", "_") + ".bpmn20.xml";
            HttpServletResponse response = ((ServletRequestAttributes) RequestContextHolder.getRequestAttributes()).getResponse();
            if (Objects.nonNull(response)) {
                response.setContentType("application/xml");
                response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
                response.setHeader("Content-Disposition", "attachment; filename=file.bpmn20.xml; filename*=" + URLEncoder.encode(fileName, "UTF-8"));
                ServletOutputStream servletOutputStream = response.getOutputStream();
                BufferedInputStream in = new BufferedInputStream(new ByteArrayInputStream(xmlBytes));
                byte[] buffer = new byte[8096];
                while (true) {
                    int count = in.read(buffer);
                    if (count == -1) {
                        break;
                    }
                    servletOutputStream.write(buffer, 0, count);
                }
                // 刷新并关闭流
                servletOutputStream.flush();
                servletOutputStream.close();
            }
        }
    }

    @Override
    public FlowModel getModel(String modelCode) {
        Optional<FlowModel> optional = flowModelDao.lambdaQueryChain()
                .eq(FlowModel::getCode, modelCode)
                .last(SqlConstant.LIMIT1)
                .oneOpt();
        return optional.orElseGet(() -> flowModelHistoryService.lambdaQuery().eq(FlowModelHistory::getCode, modelCode).one());
    }

    @Override
    public List<FlowModelHistory> getModelHistory(String modelCode) {
        return flowModelHistoryService.lambdaQuery()
                .select(FlowModelHistory.class, i -> !StringUtils.equalsAny(i.getColumn(), "process", "form", "settings"))
                .eq(FlowModelHistory::getCode, modelCode)
                .list();
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void switchModel(Long historyModelId) {
        FlowModelHistory flowModelHistory = flowModelHistoryService.getById(historyModelId);
        FlowModel flowModel = flowModelDao.selectById(flowModelHistory.getModelId());
        flowModelHistoryService.removeById(historyModelId);
        flowModelHistoryService.save(flowModel);
        flowModelDao.deleteById(flowModel);
        flowModelDao.insert(flowModelHistory);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowModel create(FlowModel flowModel) {
        flowModel.setCode(String.format("flow-%s", IdWorker.getIdStr()));
        flowModelDao.insert(flowModel);
        // 打印模板
        ReportTemplate template = new ReportTemplate();
        template.setBusinessId(flowModel.getId());
        template.setCode(flowModel.getCode());
        template.setName(flowModel.getName());
        return flowModel;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowModel update(FlowModel flowModel) {
        // 更新 FlowModel 表
        flowModelDao.updateById(flowModel);

        Long modelId = flowModel.getId();
        Long groupId = flowModel.getGroupId();

        // 查询与 FlowModel 关联的所有 FlowDefine 记录
        List<FlowDefine> flowDefineList = flowDefineService.lambdaQuery()
                .eq(FlowDefine::getModelId, modelId)
                .list();

        // 如果没有找到对应的 FlowDefine 记录，抛出异常
        if (flowDefineList.isEmpty()) {
            throw new BaseException("模版数据更新失败，流程id：" + modelId);
        }

        // 更新 FlowDefine 记录的 groupId
        flowDefineList.forEach(flowDefine -> flowDefine.setGroupId(groupId));

        // 批量更新 FlowDefine 表
        flowDefineService.updateBatchById(flowDefineList);

        return flowModel;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void delete(Set<Long> ids) {
        flowModelDao.deleteBatchIds(ids);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowDefine deploy(Long modelId) throws JsonProcessingException {
        FlowModel flowModel = flowModelDao.selectById(modelId);
        if (!flowModel.getEnable()) {
            throw new BaseException("模型已停用无法发布");
        }
        // 保存历史
        flowModelHistoryService.save(flowModel);
        flowModel.setVersion(flowModel.getVersion() + 1);
        flowModelDao.updateById(flowModel);
        BpmnModel bpmnModel = flowModel.toBpmnModel();
        FormModel formModel = flowModel.getForm();
        String formJson = objectMapper.writeValueAsString(formModel);
        Deployment deploy = repositoryService.createDeployment()
                .key(flowModel.getCode())
                .name(flowModel.getName())
                .category(flowModel.getGroupId().toString())
                .addBpmnModel(String.format("%s.bpmn20.xml", flowModel.getName()), bpmnModel)
                .addString(String.format("%s.form", flowModel.getName()), formJson)
                // .enableDuplicateFiltering()
                .deploy();
        ProcessDefinition definition = repositoryService.createProcessDefinitionQuery()
                .deploymentId(deploy.getId())
                .singleResult();
        FlowDefine flowDefine = new FlowDefine(definition);
        flowDefine.setModelId(flowModel.getId());
        flowDefine.setGroupId(flowModel.getGroupId());
        flowDefine.setProcess(flowModel.getProcess());
        flowDefine.setForm(flowModel.getForm());
        flowDefine.setIcon(flowModel.getIcon());
        flowDefine.setAdmin(flowModel.getAdmin());
        flowDefine.setStarters(flowModel.getStarters());
        flowDefine.setSettings(flowModel.getSettings());
        flowDefineService.save(flowDefine);
        return flowDefine;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowModel suspend(Long modelId) {
        FlowModel flowModel = flowModelDao.selectById(modelId);
        boolean enable = !flowModel.getEnable();
        flowModel.setEnable(enable);
        flowModelDao.updateById(flowModel);
        ProcessDefinition processDefinition = repositoryService.createProcessDefinitionQuery()
                .processDefinitionKey(flowModel.getCode())
                .latestVersion()
                .singleResult();
        if (Objects.nonNull(processDefinition)) {
            flowDefineService.suspend(processDefinition.getId());
        }
        return flowModel;
    }

    @Override
    public List<DatasetField> getReportField(String modelId) {
        FlowModel flowModel = flowModelDao.selectById(modelId);
        if (Objects.isNull(flowModel)) {
            throw new BaseException("流程不存在");
        }
        FormModel formModel = flowModel.getForm();
        Map<String, Field> fieldMap = formModel.allFieldsAsMap();
        List<DatasetField> list = fieldMap.values().stream().map(field -> new DatasetField(
                field.getLabel(),
                String.format("${%s}", field.getId())
        )).collect(Collectors.toList());
        list.add(new DatasetField("发起人", "${_initiator_}"));
        list.add(new DatasetField("编号", "${_code_}"));
        list.add(new DatasetField("标题", "${_title_}"));
        list.add(new DatasetField("状态", "${_status_}"));
        list.add(new DatasetField("发起时间", "${_startTime_}"));
        list.add(new DatasetField("结束时间", "${_endTime_}"));
        list.add(new DatasetField("打印时间", "${_currentTime_}"));
        DatasetField datasetField = new DatasetField();
        datasetField.setField(String.format("${%s}", modelId));
        datasetField.setLabel(flowModel.getName());
        datasetField.setChildren(list);
        return Lists.newArrayList(datasetField);
    }

    @Override
    public SpreadSheet savePrint(SpreadSheet spreadSheet, String modelId) {
        FlowModel flowModel = flowModelDao.selectById(modelId);
        if (Objects.isNull(flowModel)) {
            throw new BaseException("流程模型不存在");
        }
        Settings settings = flowModel.getSettings();
        PrintConfig printConfig = settings.getPrint();
        printConfig.setTemplate(spreadSheet);
        flowModel.setSettings(settings);
        flowModelDao.updateById(flowModel);
        return spreadSheet;
    }

    @Override
    public SpreadSheet getPrint(String modelId) {
        FlowModel flowModel = flowModelDao.selectById(modelId);
        Settings settings = flowModel.getSettings();
        PrintConfig printConfig = settings.getPrint();
        return printConfig.getTemplate();
    }

    @Override
    public void exportExcel(String modelId) throws UnsupportedEncodingException {
        SpreadSheet spreadSheet = this.getPrint(modelId);
        SXSSFWorkbook excel = spreadSheet.toExcel();
        // 响应前端
        HttpServletResponse response = ((ServletRequestAttributes) RequestContextHolder.getRequestAttributes()).getResponse();
        response.setContentType("application/vnd.ms-excel");
        response.setCharacterEncoding("utf-8");
        response.setHeader("Content-disposition", "attachment;filename=" + URLEncoder.encode(spreadSheet.getName() + ".xlsx", "UTF-8"));
        try {
            ServletOutputStream outputStream = response.getOutputStream();
            excel.write(outputStream);
            outputStream.flush();
            outputStream.close();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
}
